==> cog_suspect-1.5.2-py3-none-any.whl: suspect/summary.py <==
# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
import logging
from pyomo.core.kernel.component_map import ComponentMap
from suspect.pyomo.convert import dag_from_pyomo_model
from suspect.fbbt import perform_fbbt
from suspect.dag.iterator import DagForwardIterator, DagBackwardIterator
from suspect.context import SpecialStructurePropagationContext
from suspect.interval import Interval
from suspect.propagation import propagate_special_structure
from suspect.polynomial_degree import polynomial_degree
from pyomo.environ import ConcreteModel # pylint: disable=no-name-in-module,import-error


logger = logging.getLogger('suspect')


class ModelInformation(object):
    """Hold structure information about a problem.

    Attributes
    ----------
    variables : dict
       map variables names to their information
    objectives : dict
       map objectives names to their information
    constraints : dict
       map constraints names to their information
    """
    def __init__(self, name, variables, objectives, constraints):
        self.name = name
        self.variables = variables
        self.objectives = objectives
        self.constraints = constraints

    def num_variables(self):
        """Returns the number of variables in the problem."""
        return len(self.variables)

    def num_binaries(self):
        """Returns the number of binary variables in the problem."""
        binaries = [
            v for v in self.variables.values()
            if v['type'] == 'binary'
        ]
        return len(binaries)

    def num_integers(self):
        """Returns the number of integer variables in the problem."""
        integers = [
            v for v in self.variables.values()
            if v['type'] == 'integer'
        ]
        return len(integers)

    def num_constraints(self):
        """Returns the number of constraints in the problem."""
        return len(self.constraints)

    def conscurvature(self):
        """Returns the convexity of each constraint."""
        return dict(
            (k, v['convexity'])
            for k, v in self.constraints.items()
        )

    def objcurvature(self):
        """Returns the convexity of each objective."""
        return dict(
            (k, v['convexity'])
            for k, v in self.objectives.items()
        )

    def objtype(self):
        """Returns the type of each objective."""
        def _objtype(v):
            cvx = v['convexity']
            deg = v['polynomial_degree']
            if cvx.is_linear():
                assert deg.is_linear()
                return 'linear'
            elif deg.is_quadratic():
                return 'quadratic'
            elif deg.is_polynomial():
                return 'polynomial'
            else:
                return 'nonlinear'

        return dict(
            (k, _objtype(v))
            for k, v in self.objectives.items()
        )


def detect_special_structure(problem, max_iter=10):
    """Detect special structure in the problem.

    Parameters
    ----------
    model : suspect.dag.ProblemDag or ConcreteModel
        the problem DAG or a Pyomo ConcreteModel.

    max_iter : int
        number of maximum bound propagation/tightening iterations.

    Returns
    -------
    ModelInformation
        an object containing the detected infomation about the problem.
    """
    if isinstance(problem, ConcreteModel):
        problem = dag_from_pyomo_model(problem)

    bounds = perform_fbbt(problem)

    polynomial = polynomial_degree(problem)
    monotonicity, convexity = propagate_special_structure(problem, bounds)

    variables = {}
    for variable_name, variable in problem.variables.items():
        if variable.is_binary():
            variable_type = 'binary'
        elif variable.is_integer():
            variable_type = 'integer'
        else:
            variable_type = 'continuous'
        var_bounds = bounds[variable]

        if variable.name in variables:
            warnings.warn('Duplicate variable {}'.format(variable.name))

        variables[variable_name] = {
            'name': variable_name,
            'type': variable_type,
            'lower_bound': var_bounds.lower_bound,
            'upper_bound': var_bounds.upper_bound,
        }

    objectives = {}
    for obj_name, obj in problem.objectives.items():
        if obj_name in objectives:
            warnings.warn('Duplicate objective {}'.format(obj_name))

        if obj.is_minimizing():
            sense = 'min'
        else:
            sense = 'max'
        obj_bounds = bounds.get(obj, Interval(None, None))
        cvx = convexity[obj]
        poly = polynomial[obj]

        objectives[obj.name] = {
            'sense': sense,
            'convexity': cvx,
            'polynomial_degree': poly,
            'lower_bound': obj_bounds.lower_bound,
            'upper_bound': obj_bounds.upper_bound,
        }

    constraints = {}
    for cons_name, cons in problem.constraints.items():
        if cons_name in constraints:
            warnings.warn('Duplicate constraint {}'.format(cons_name))

        if cons.is_equality():
            type_ = 'equality'
        else:
            type_ = 'inequality'

        cons_bounds = bounds.get(cons, Interval(None, None))
        cvx = convexity[cons]
        poly = polynomial[cons]

        constraints[cons_name] = {
            'type': type_,
            'convexity': cvx,
            'polynomial_degree': poly,
            'lower_bound': cons_bounds.lower_bound,
            'upper_bound': cons_bounds.upper_bound,
        }

    return ModelInformation(
        name=problem.name,
        variables=variables,
        objectives=objectives,
        constraints=constraints,
    )

==> cog_suspect-2.0.1-py3-none-any.whl: suspect/summary.py <==
# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import warnings

import pyomo.environ as pe

from suspect.fbbt import perform_fbbt
from suspect.interval import Interval
from suspect.convexity import Convexity
from suspect.polynomial_degree import polynomial_degree
from suspect.propagation import propagate_special_structure

logger = logging.getLogger('suspect')


class ModelInformation(object):
    """Hold structure information about a problem.

    Attributes
    ----------
    variables : dict
       map variables names to their information
    objectives : dict
       map objectives names to their information
    constraints : dict
       map constraints names to their information
    """
    def __init__(self, name, variables, objectives, constraints):
        self.name = name
        self.variables = variables
        self.objectives = objectives
        self.constraints = constraints

    def num_variables(self):
        """Returns the number of variables in the problem."""
        return len(self.variables)

    def num_binaries(self):
        """Returns the number of binary variables in the problem."""
        binaries = [
            v for v in self.variables.values()
            if v['type'] == 'binary'
        ]
        return len(binaries)

    def num_integers(self):
        """Returns the number of integer variables in the problem."""
        integers = [
            v for v in self.variables.values()
            if v['type'] == 'integer'
        ]
        return len(integers)

    def num_constraints(self):
        """Returns the number of constraints in the problem."""
        return len(self.constraints)

    def conscurvature(self):
        """Returns the convexity of each constraint."""
        return dict(
            (k, v['convexity'])
            for k, v in self.constraints.items()
        )

    def objcurvature(self):
        """Returns the convexity of each objective."""
        return dict(
            (k, v['convexity'])
            for k, v in self.objectives.items()
        )

    def objtype(self):
        """Returns the type of each objective."""
        def _objtype(v):
            cvx = v['convexity']
            deg = v['polynomial_degree']
            if cvx.is_linear():
                assert deg == 1
                return 'linear'
            elif deg == 2:
                return 'quadratic'
            elif deg is not None:
                return 'polynomial'
            else:
                return 'nonlinear'

        return dict(
            (k, _objtype(v))
            for k, v in self.objectives.items()
        )


def detect_special_structure(problem, max_iter=10):
    """Detect special structure in the problem.

    Parameters
    ----------
    model : suspect.dag.ProblemDag or ConcreteModel
        the problem DAG or a Pyomo ConcreteModel.

    max_iter : int
        number of maximum bound propagation/tightening iterations.

    Returns
    -------
    ModelInformation
        an object containing the detected infomation about the problem.
    """
    bounds = perform_fbbt(problem)

    monotonicity, convexity = propagate_special_structure(problem, bounds)

    variables = {}
    for variable in problem.component_data_objects(pe.Var, active=True, descend_into=True):
        if variable.is_binary():
            variable_type = 'binary'
        elif variable.is_integer():
            variable_type = 'integer'
        else:
            variable_type = 'continuous'
        var_bounds = bounds[variable]
        variable_name = variable.name

        variables[variable_name] = {
            'name': variable_name,
            'type': variable_type,
            'lower_bound': var_bounds.lower_bound,
            'upper_bound': var_bounds.upper_bound,
        }

    objectives = {}
    for obj in problem.component_data_objects(pe.Objective, active=True, descend_into=True):
        if obj.is_minimizing():
            sense = 'min'
        else:
            sense = 'max'
        obj_bounds = bounds.get(obj.expr, Interval(None, None))
        cvx = convexity[obj.expr]
        poly = obj.expr.polynomial_degree()

        if sense == 'max':
            cvx = cvx.negate()

        objectives[obj.name] = {
            'sense': sense,
            'convexity': cvx,
            'polynomial_degree': poly,
            'lower_bound': obj_bounds.lower_bound,
            'upper_bound': obj_bounds.upper_bound,
        }

    constraints = {}
    for cons in problem.component_data_objects(pe.Constraint, active=True, descend_into=True):
        if cons.has_lb() and cons.has_ub():
            type_ = 'equality'
        else:
            type_ = 'inequality'

        cons_bounds = bounds.get(cons.body, Interval(None, None))
        cvx = convexity[cons.body]
        poly = cons.body.polynomial_degree()

        if type_ == 'equality':
            if not cvx.is_linear():
                cvx = Convexity.Unknown
        else:
            if cons.has_lb():
                cvx = cvx.negate()

        constraints[cons.name] = {
            'type': type_,
            'convexity': cvx,
            'polynomial_degree': poly,
            'lower_bound': cons_bounds.lower_bound,
            'upper_bound': cons_bounds.upper_bound,
        }

    return ModelInformation(
        name=problem.name,
        variables=variables,
        objectives=objectives,
        constraints=constraints,
    )
